@@ -1,5 +1,5 @@
 /* Texas Instruments OMAP framebuffer driver for X.Org
- * Copyright 2008, 2010 Kalle Vahlman, <zuh@iki.fi>
+ * Copyright 2008 Kalle Vahlman, <zuh@iki.fi>
  *
  * The driver setup in this file is adapted from the fbdev driver
  * Original authors of the fbdev driver:
@@ -36,20 +36,13 @@
 #include "xf86.h"
 #include "xf86_OSlib.h"
 
-#include "xf86Crtc.h"
-
 #include "micmap.h"
-#include "mipointer.h"
 #include "fb.h"
 
 #include "exa.h"
 
-#ifdef HAVE_XEXTPROTO_71
-#include <X11/extensions/dpmsconst.h>
-#else
 #define DPMS_SERVER
 #include <X11/extensions/dpms.h>
-#endif
 
 #include <linux/fb.h>
 
@@ -59,9 +52,6 @@
 #include "omapfb.h"
 
 #include "omapfb-driver.h"
-#include "omapfb-crtc.h"
-#include "omapfb-output.h"
-#include "omapfb-utils.h"
 
 #define OMAPFB_VERSION 1000
 #define OMAPFB_DRIVER_NAME "OMAPFB"
@@ -72,7 +62,12 @@ static Bool OMAPFBPreInit(ScrnInfoPtr pScrn, int flags);
 static Bool OMAPFBScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv);
 static Bool OMAPFBEnterVT(int scrnIndex, int flags);
 static void OMAPFBLeaveVT(int scrnIndex, int flags);
+static Bool OMAPFBSaveScreen(ScreenPtr pScreen, int mode);
 static Bool OMAPFBSwitchMode(int scrnIndex, DisplayModePtr mode, int flags);
+static void OMAPFBDPMSSet(ScrnInfoPtr pScrn, int mode, int flags);
+
+static void setup_default_mode(OMAPFBPtr ofb);
+static Bool set_mode(OMAPFBPtr ofb, DisplayModePtr mode);
 
 static Bool
 OMAPFBEnsureRec(ScrnInfoPtr pScrn)
@@ -89,7 +84,7 @@ OMAPFBFreeRec(ScrnInfoPtr pScrn)
 {
 	if (pScrn->driverPrivate == NULL)
 		return;
-	free(pScrn->driverPrivate);
+	xfree(pScrn->driverPrivate);
 	pScrn->driverPrivate = NULL;
 }
 
@@ -129,6 +124,26 @@ OMAPFBIdentify(int flags)
 	                  OMAPFBChipsets);
 }
 
+static const char *fbSymbols[] = {
+	"fbScreenInit",
+	"fbPictureInit",
+	NULL
+};
+
+static const char *exaSymbols[] = {
+    "exaDriverAlloc",
+    "exaDriverInit",
+    "exaDriverFini",
+    "exaOffscreenAlloc",
+    "exaOffscreenFree",
+    "exaGetPixmapOffset",
+    "exaGetPixmapPitch",
+    "exaGetPixmapSize",
+    "exaMarkSync",
+    "exaWaitSync",
+    NULL
+};
+
 static void
 OMAPFBProbeController(char *ctrl_name)
 {
@@ -145,7 +160,8 @@ OMAPFBProbeController(char *ctrl_name)
 		        SYSFS_LCTRL_FILE, strerror(errno));
 	} else {
 		int s = read(fd, ctrl_name, 31);
-		if (s > 0) {
+		if (s > 0)
+		{
 			ctrl_name[s-1] = '\0';
 			found = TRUE;
 		} else {
@@ -171,62 +187,38 @@ OMAPFBProbe(DriverPtr drv, int flags)
 	int i;
        	GDevPtr *devSections;
 	int numDevSections;
-	char *config_dev;
+	int bus,device,func;
+	char *dev;
 	ScrnInfoPtr pScrn = NULL;
 	Bool foundScreen = FALSE;
-	int fd;
-	int entity;
-	struct fb_fix_screeninfo info;
-	OMAPFBPtr ofb;
 
 	if (flags & PROBE_DETECT) return FALSE;
 
 	/* Search for device sections for us */
-	if ((numDevSections = xf86MatchDevice(OMAPFB_NAME, &devSections)) <= 0) {
-		xf86Msg(X_ERROR,
-		        "%s: Found 0 device sections.\n", __FUNCTION__);
+	if ((numDevSections = xf86MatchDevice(OMAPFB_NAME, &devSections)) <= 0) 
 		return FALSE;
-	}
 
-	if (numDevSections > 1) {
-		xf86Msg(X_WARNING,
-		        "%s: Found %d device sections but supporting only one."
-		        "Ignoring the rest.\n", __FUNCTION__, numDevSections);
-	}
+/* FIXME: We don't really want to do it like this... */
+#define DEFAULT_DEVICE "/dev/fb"
 
-	/* Fetch the device path */
-	config_dev = xf86FindOptionValue(devSections[0]->options, "fb");
-
-#define FB_MAX_FD 8
-	for (i = -1; i < FB_MAX_FD; i++) {
-		char device_path[PATH_MAX];
-
-		/* Concatenate the iterative device path (first attempt
-		   without a number) */
-		if (i == -1) {
-			snprintf(device_path, PATH_MAX, "/dev/fb");
-		} else {
-			snprintf(device_path, PATH_MAX, "/dev/fb%d", i);
-		}
-
-		/* Try opening it to see if we can access it. Use the device path
-		   from configuration, if set */
-		fd = open(config_dev != NULL ? config_dev : device_path, O_RDWR, 0);
+	for (i = 0; i < numDevSections; i++) {
+		int fd;
+		
+		/* Fetch the device path */
+		dev = xf86FindOptionValue(devSections[i]->options, "fb");
 
+		/* Try opening it to see if we can access it */
+		fd = open(dev != NULL ? dev : DEFAULT_DEVICE, O_RDWR, 0);
 		if (fd > 0) {
+			int entity;
+			struct fb_fix_screeninfo info;
+
 			if (ioctl (fd, FBIOGET_FSCREENINFO, &info)) {
 				xf86Msg(X_WARNING,
 				        "%s: Reading hardware info failed: %s\n",
 				        __FUNCTION__, strerror(errno));
 				close(fd);
-
-				/* If using a configured device, return an
-				   error. Otherwise continue iteration. */
-				if (config_dev != NULL) {
-					return FALSE;
-				} else {
-					continue;
-				}
+				continue;
 			}
 			close(fd);
 
@@ -245,7 +237,7 @@ OMAPFBProbe(DriverPtr drv, int flags)
 			foundScreen = TRUE;
 
 			/* Tell the rest of the drivers that this one is ours */
-			entity = xf86ClaimFbSlot(drv, 0, devSections[0], TRUE);
+			entity = xf86ClaimFbSlot(drv, 0, devSections[i], TRUE);
 			pScrn = xf86ConfigFbEntity(pScrn, 0, entity,
 			                           NULL, NULL, NULL, NULL);
 
@@ -259,24 +251,14 @@ OMAPFBProbe(DriverPtr drv, int flags)
 			pScrn->EnterVT       = OMAPFBEnterVT;
 			pScrn->LeaveVT       = OMAPFBLeaveVT;
 
-			/* Get our private data */
-			OMAPFBEnsureRec(pScrn);
-			ofb = OMAPFB(pScrn);
-
-			/* Store the path to framebuffer device */
-			strncpy(ofb->fb_path, config_dev ? config_dev : device_path, PATH_MAX);
-
-			/* Succeeded in finding the framebuffer. */
-			break;
-
 		} else {
-			xf86Msg(X_WARNING, "Could not open '%s': %s\n",
-			        config_dev ? config_dev : device_path, strerror(errno));
+			xf86Msg(X_WARNING, "Could not open '%s': %s",
+			        dev ? dev : DEFAULT_DEVICE, strerror(errno));
 		}
 
 	}
 
-	free(devSections);
+	xfree(devSections);
 
 	return foundScreen;
 }
@@ -286,8 +268,11 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 {
 	OMAPFBPtr ofb;
 	EntityInfoPtr pEnt;
+	int fd;
+	char *dev;
 	rgb zeros = { 0, 0, 0 };
-	struct stat st;
+	struct omapfb_caps caps;
+	char ctrl_name[32];
 
 	if (flags & PROBE_DETECT) return FALSE;
 	
@@ -305,16 +290,19 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 	pEnt = xf86GetEntityInfo(pScrn->entityList[0]);
 	
 	/* Open the device node */
-	ofb->fd = open(ofb->fb_path, O_RDWR, 0);
-	if (ofb->fd == -1) {
+	dev = xf86FindOptionValue(pEnt->device->options, "fb");
+	ofb->fd = open(dev != NULL ? dev : DEFAULT_DEVICE, O_RDWR, 0);
+	if (ofb->fd == -1)
+	{
 		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		           "%s: Opening '%s' failed: %s\n", __FUNCTION__,
-		           ofb->fb_path, strerror(errno));
+		           dev != NULL ? dev : DEFAULT_DEVICE, strerror(errno));
 		OMAPFBFreeRec(pScrn);
 		return FALSE;
 	}
 
-	if (ioctl (ofb->fd, FBIOGET_FSCREENINFO, &ofb->fixed_info)) {
+	if (ioctl (ofb->fd, FBIOGET_FSCREENINFO, &ofb->fixed_info))
+	{
 		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		           "%s: Reading hardware info failed: %s\n",
 		           __FUNCTION__, strerror(errno));
@@ -325,21 +313,15 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 	/* Try to detect what LCD controller we're using */
 	OMAPFBProbeController(ofb->ctrl_name);
 
-	/* Do we have the DSS kernel API? */
-	if (stat(SYSFS_DSS_DIR, &st) == 0) {
-		ofb->dss = TRUE;
-	} else {
-		ofb->dss = FALSE;
-	}
-
 	/* Print out capabilities, if available */
-	if (!ioctl (ofb->fd, OMAPFB_GET_CAPS, &ofb->caps)) {
-		OMAPFBPrintCapabilities(pScrn, &ofb->caps,
+	if (!ioctl (fd, OMAPFB_GET_CAPS, &caps)) {
+		OMAPFBPrintCapabilities(pScrn, &caps,
 		                        "Base plane");
 	}
 
 	/* Check the memory setup. */
-	if (ioctl (ofb->fd, OMAPFB_QUERY_MEM, &ofb->mem_info)) {
+	if (ioctl (ofb->fd, OMAPFB_QUERY_MEM, &ofb->mem_info))
+	{
 		/* As a fallback, set up the mem_info struct from info we know */
 		ofb->mem_info.type = OMAPFB_MEMTYPE_SDRAM;
 		ofb->mem_info.size = ofb->fixed_info.smem_len;
@@ -350,7 +332,8 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 	           pScrn->videoRam/1024,
 	           ofb->mem_info.type == OMAPFB_MEMTYPE_SDRAM ? "SDRAM" : "SRAM");
 
-	if (ioctl (ofb->fd, FBIOGET_VSCREENINFO, &ofb->state_info)) {
+	if (ioctl (ofb->fd, FBIOGET_VSCREENINFO, &ofb->state_info))
+	{
 		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		           "%s: Reading screen state info failed: %s\n",
 		           __FUNCTION__, strerror(errno));
@@ -362,33 +345,38 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 			     0, /* depth: overall server default */
 			     ofb->state_info.bits_per_pixel,
 			     ofb->state_info.bits_per_pixel,
-			     Support24bppFb | Support32bppFb))
+			     0))
 		return FALSE;
 
 	xf86PrintDepthBpp(pScrn);
 
 	/* This apparently sets the color weights. We're feeding it zeros. */
 	if (!xf86SetWeight(pScrn, zeros, zeros)) {
-		return FALSE;
-	}
+            return FALSE;
+        }
 
 	/* Initialize default visual */
 	if (!xf86SetDefaultVisual(pScrn, -1))
 		return FALSE;
 
+/* FIXME: We should allow options for things like overlay framebuffers,
+          rotation, etc
+	xf86CollectOptions(pScrn, NULL);
+*/
+
 	pScrn->progClock = TRUE;
 	pScrn->chipset   = "omapfb";
 	
 	/* Start with configured virtual size */
 	pScrn->virtualX = pScrn->display->virtualX;
 	pScrn->virtualY = pScrn->display->virtualY;
-	pScrn->displayWidth = ofb->fixed_info.line_length / (ofb->state_info.bits_per_pixel >> 3);
+	pScrn->displayWidth = ofb->state_info.xres;
 
-	/* Clamp to actual virtual resolution */
-	if (pScrn->virtualX < ofb->state_info.xres_virtual)
-		pScrn->virtualX = ofb->state_info.xres_virtual;
-	if (pScrn->virtualY < ofb->state_info.yres_virtual)
-		pScrn->virtualY = ofb->state_info.yres_virtual;
+	/* Clamp to actual resolution */
+	if (pScrn->virtualX < ofb->state_info.xres)
+		pScrn->virtualX = ofb->state_info.xres;
+	if (pScrn->virtualY < ofb->state_info.yres)
+		pScrn->virtualY = ofb->state_info.yres;
 	
 	/* Setup viewport */
 	pScrn->frameX0 = 0;
@@ -396,8 +384,20 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 	pScrn->frameX1 = ofb->state_info.xres;
 	pScrn->frameY1 = ofb->state_info.yres;
 
-	pScrn->maxVValue = ofb->state_info.xres_virtual;
-	pScrn->maxHValue = ofb->state_info.yres_virtual;
+	pScrn->maxVValue = ofb->state_info.xres;
+	pScrn->maxHValue = ofb->state_info.yres;
+
+	/* Setup default mode as the mode the fb is in on startup, *usually*
+	 * it'll be the one we want anyway
+	 */
+	setup_default_mode(ofb);
+
+	/* TODO: Grab all provided modes from config and validate them */
+	ofb->default_mode.status = xf86CheckModeForMonitor(&ofb->default_mode, pScrn->monitor);
+
+	/* TODO: Add all validated modes here */
+	pScrn->modes = &ofb->default_mode;
+	pScrn->currentMode = pScrn->modes;
 
 	pScrn->offset.red   = ofb->state_info.red.offset;
 	pScrn->offset.green = ofb->state_info.green.offset;
@@ -408,30 +408,7 @@ OMAPFBPreInit(ScrnInfoPtr pScrn, int flags)
 	                          << ofb->state_info.green.offset;
 	pScrn->mask.blue    = ((1 << ofb->state_info.blue.length) - 1)
 	                          << ofb->state_info.blue.offset;
-
-	pScrn->modes = NULL;
-
-	OMAPFBCRTCInit(pScrn);
-
-	if (ofb->dss) {
-		OMAPFBOutputInitDSS(pScrn);
-	} else {
-		OMAPFBOutputInit(pScrn);
-	}
-
-	if (xf86InitialConfiguration(pScrn, TRUE)) {
-		xf86DrvMsg(pScrn->scrnIndex, X_INFO, "XRandR extension initialized\n");
-	} else {
-		xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "Initial output configuration failed!\n");
-	}
-
-	pScrn->currentMode = pScrn->modes;
-	ofb->crtc->mode = *pScrn->currentMode;
-
-
-	/* Disable outputs that are not used */
-	xf86DisableUnusedFunctions(pScrn);
-
+	
 	xf86PrintModes(pScrn);
 	
 	/* Set the screen dpi value (we don't give defaults) */
@@ -457,7 +434,7 @@ OMAPFBXvScreenInit(ScreenPtr pScreen)
 	if (n > 0 || on > 0) {
 		int i;
 		XF86VideoAdaptorPtr *generic_adaptors = ptr;
-		ptr = malloc((n + on) * sizeof(XF86VideoAdaptorPtr));
+		ptr = xalloc((n + on) * sizeof(XF86VideoAdaptorPtr));
 		for (i = 0; i < n; i++) {
 			ptr[i] = generic_adaptors[i];
 		}
@@ -482,9 +459,9 @@ OMAPFBCloseScreen(int scrnIndex, ScreenPtr pScreen)
 
 	munmap(ofb->fb, ofb->mem_info.size);
 
-	pScreen->CloseScreen = ofb->CloseScreen;
+        pScreen->CloseScreen = ofb->CloseScreen;
 	
-	return (*pScreen->CloseScreen)(scrnIndex, pScreen);
+        return (*pScreen->CloseScreen)(scrnIndex, pScreen);
 }
 
 static Bool
@@ -525,6 +502,7 @@ OMAPFBScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 	/* Load the fallback module */
 	xf86LoadSubModule(pScrn, "fb");
+	xf86LoaderReqSymLists(fbSymbols, NULL);
 
 	/* Initialize fallbacks for the screen */
 	if (!fbScreenInit(pScreen, ofb->fb, pScrn->virtualX,
@@ -548,7 +526,7 @@ OMAPFBScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 				visual->blueMask = pScrn->mask.blue;
 			}
 		}
-	}
+	   }
 
 	/* Initialize XRender fallbacks */
 	if (!fbPictureInit(pScreen, NULL, 0)) {
@@ -569,19 +547,24 @@ OMAPFBScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		return FALSE;
 	}
 
+	/* Enforce the default mode (this is silly I guess) */
+	set_mode(ofb, &ofb->default_mode);
+
 	/* Make sure the plane is up and running */
-	if (ioctl (ofb->fd, OMAPFB_QUERY_PLANE, &ofb->plane_info)) {
+	if (ioctl (ofb->fd, OMAPFB_QUERY_PLANE, &ofb->plane_info))
+	{
 		/* This is non-fatal since we might be running against older
 		 * kernel driver in which case we only do basic 2D stuff...
 		 */
 		xf86DrvMsg(scrnIndex, X_ERROR, "Reading plane info failed\n");
-	} else if (!ofb->dss) {
+	} else {
 
 		ofb->plane_info.enabled = 1;
 		ofb->plane_info.out_width = ofb->state_info.xres;
 		ofb->plane_info.out_height = ofb->state_info.yres;
 
-		if (ioctl (ofb->fd, OMAPFB_SETUP_PLANE, &ofb->plane_info)) {
+		if (ioctl (ofb->fd, OMAPFB_SETUP_PLANE, &ofb->plane_info))
+		{
 			xf86DrvMsg(scrnIndex, X_ERROR,
 			            "%s: Plane setup failed: %s\n",
 			            __FUNCTION__, strerror(errno));
@@ -594,40 +577,182 @@ OMAPFBScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		           "FBIOBLANK: %s\n", strerror(errno));
 	}
 
-	/* Setup screen saving and DPMS support, these will go through the
-	 * Output functions
-	 */
-	pScreen->SaveScreen = xf86SaveScreen;
-	xf86DPMSInit(pScreen, xf86DPMSSet, 0);
-
+	/* NO-OP currently */
+	pScreen->SaveScreen = OMAPFBSaveScreen;
+	
+	/* Setup DPMS support */
+	xf86DPMSInit(pScreen, OMAPFBDPMSSet, 0);
+	
 #ifdef USE_EXA
 	/* EXA init */
 	xf86LoadSubModule(pScrn, "exa");
+	xf86LoaderReqSymLists(exaSymbols, NULL);
 
 	/* TODO: This should depend on the AccelMethod option */
 	ofb->exa = exaDriverAlloc();
-	if (OMAPFBSetupExa(ofb)) {
+	if (OMAPFBSetupExa(ofb))
+	{
 		exaDriverInit(pScreen, ofb->exa);
 	} else {
-		free(ofb->exa);
+		xfree(ofb->exa);
 		ofb->exa = NULL;
 	}
 #endif
 
 	/* Initialize XVideo support */
-	/* FIXME: Currently dss & XV do not co-exist */
-	if (!ofb->dss)
-		OMAPFBXvScreenInit(pScreen);
+	OMAPFBXvScreenInit(pScreen);
+	
+	/* TODO: RANDR support */
 
-	/* Initialize RANDR support */
-	xf86CrtcScreenInit(pScreen);
+	
+	return TRUE;
+}
+
+/* Adapted from fbdevhw.c in X.Org server sources */
+static Bool
+set_mode(OMAPFBPtr ofb, DisplayModePtr mode)
+{
+#ifdef ENFORCE_MODES
+	struct fb_var_screeninfo var;
+	
+	var = ofb->state_info;
+
+	var.xres = mode->HDisplay;
+	var.yres = mode->VDisplay;
+	if (var.xres_virtual < var.xres)
+		var.xres_virtual = var.xres;
+	if (var.yres_virtual < var.yres)
+		var.yres_virtual = var.yres;
+	var.xoffset = var.yoffset = 0;
+	var.pixclock = mode->Clock ? 1000000000/mode->Clock : 0;
+	var.right_margin = mode->HSyncStart - mode->HDisplay;
+	var.hsync_len = mode->HSyncEnd - mode->HSyncStart;
+	var.left_margin = mode->HTotal - mode->HSyncEnd;
+	var.lower_margin = mode->VSyncStart - mode->VDisplay;
+	var.vsync_len = mode->VSyncEnd - mode->VSyncStart;
+	var.upper_margin = mode->VTotal - mode->VSyncEnd;
+	var.sync = 0;
+	var.bits_per_pixel = 16;
+	var.grayscale = 0;
+	var.nonstd = 0;
+	var.accel_flags = 0;
+	var.activate = FB_ACTIVATE_NOW;
+	var.rotate = FB_ROTATE_UR;
+	var.vmode = FB_VMODE_NONINTERLACED;
+
+	if (ioctl (ofb->fd, FBIOPUT_VSCREENINFO, &var))
+	{
+		return FALSE;
+	}
+
+	if (ioctl (ofb->fd, FBIOGET_VSCREENINFO, &ofb->state_info))
+	{
+		return FALSE;
+	}
+#endif
 
 	return TRUE;
 }
 
+/* Adapted from fbdevhw.c in X.Org server sources */
+static void
+setup_default_mode(OMAPFBPtr ofb)
+{
+	ofb->default_mode.name = "current";
+	ofb->default_mode.next = &ofb->default_mode;
+	ofb->default_mode.prev = &ofb->default_mode;
+	ofb->default_mode.type |= M_T_BUILTIN;
+
+	ofb->default_mode.HDisplay = ofb->state_info.xres;
+	ofb->default_mode.HSyncStart =   ofb->default_mode.HDisplay
+	                               + ofb->state_info.right_margin;
+	ofb->default_mode.HSyncEnd =   ofb->default_mode.HSyncStart
+	                             + ofb->state_info.hsync_len;
+	ofb->default_mode.HTotal =   ofb->default_mode.HSyncEnd
+	                           + ofb->state_info.left_margin;
+	ofb->default_mode.VDisplay = ofb->state_info.yres;
+	ofb->default_mode.VSyncStart =   ofb->default_mode.VDisplay
+	                               + ofb->state_info.lower_margin;
+	ofb->default_mode.VSyncEnd =   ofb->default_mode.VSyncStart
+	                             + ofb->state_info.vsync_len;
+	ofb->default_mode.VTotal =   ofb->default_mode.VSyncEnd
+	                           + ofb->state_info.upper_margin;
+	ofb->default_mode.SynthClock = ofb->state_info.pixclock
+	                               ? 1000000000 / ofb->state_info.pixclock
+	                               : 28000000;
+	ofb->default_mode.CrtcHDisplay = ofb->default_mode.HDisplay;
+	ofb->default_mode.CrtcHSyncStart = ofb->default_mode.HSyncStart;
+	ofb->default_mode.CrtcHSyncEnd = ofb->default_mode.HSyncEnd;
+	ofb->default_mode.CrtcHTotal = ofb->default_mode.HTotal;
+	ofb->default_mode.CrtcVDisplay = ofb->default_mode.VDisplay;
+	ofb->default_mode.CrtcVSyncStart = ofb->default_mode.VSyncStart;
+	ofb->default_mode.CrtcVSyncEnd = ofb->default_mode.VSyncEnd;
+	ofb->default_mode.CrtcVTotal = ofb->default_mode.VTotal;
+	ofb->default_mode.CrtcHAdjusted = FALSE;
+	ofb->default_mode.CrtcVAdjusted = FALSE;
+}
+
 static Bool OMAPFBSwitchMode(int scrnIndex, DisplayModePtr mode, int flags)
 {
-	return xf86SetSingleMode (xf86Screens[scrnIndex], mode, RR_Rotate_0);
+	ScrnInfoPtr pScrn = xf86Screens[scrnIndex];
+	OMAPFBPtr ofb = OMAPFB(pScrn);
+
+	if (!set_mode(ofb, mode))
+	{
+		xf86DrvMsg(scrnIndex, X_ERROR,
+		           "%s: Setting display mode failed: %s\n",
+		           __FUNCTION__, strerror(errno));
+		/* Restore the default mode as a fallback
+		 * TODO: should we do this or does X do it for us?
+		 */
+		set_mode(ofb, &ofb->default_mode);
+	}
+
+	if (ioctl (ofb->fd, FBIOGET_VSCREENINFO, &ofb->state_info))
+	{
+		xf86DrvMsg(scrnIndex, X_ERROR,
+		           "%s: Reading screen state info failed: %s\n",
+		           __FUNCTION__, strerror(errno));
+		return FALSE;
+	}
+
+	/* TODO: If we support scaling, setup the update window to do it
+	 * This needs the manual updates mode to be done.
+	 */
+	
+	return TRUE;
+}
+
+static void
+OMAPFBDPMSSet(ScrnInfoPtr pScrn, int mode, int flags)
+{
+	OMAPFBPtr ofb = OMAPFB(pScrn);
+
+	switch (mode) {
+		case DPMSModeOn:
+			if (ioctl(ofb->fd, FBIOBLANK, (void *)VESA_NO_BLANKING)) {
+				xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+				           "FBIOBLANK: %s\n", strerror(errno));
+			}
+			break;
+		case DPMSModeStandby:
+		case DPMSModeSuspend:
+			/* TODO: Maybe we would want to use the above modes for
+			 * dimming the LCD? That'd match the functionality
+			 * (save power)
+			 */
+		case DPMSModeOff:
+			/* OMAPFB only supports on and off */
+			if (ioctl(ofb->fd, FBIOBLANK, (void *)VESA_POWERDOWN)) {
+				xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+				           "FBIOBLANK: %s\n", strerror(errno));
+			}
+			break;
+			break;
+		default:
+			return;
+	}
+
 }
 
 void
@@ -682,7 +807,7 @@ static Bool
 OMAPFBEnterVT(int scrnIndex, int flags)
 {
 	xf86Msg(X_NOT_IMPLEMENTED, "%s\n", __FUNCTION__);
-	return TRUE;
+	return FALSE;
 }
 
 static void
@@ -691,19 +816,26 @@ OMAPFBLeaveVT(int scrnIndex, int flags)
 	xf86Msg(X_NOT_IMPLEMENTED, "%s\n", __FUNCTION__);
 }
 
+static Bool
+OMAPFBSaveScreen(ScreenPtr pScreen, int mode)
+{
+	xf86Msg(X_NOT_IMPLEMENTED, "%s: Dim backlight?\n", __FUNCTION__);
+	return TRUE;
+}
+
 static Bool
 OMAPFBDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op, pointer ptr)
 {
-	xorgHWFlags *flag;
-
-	switch (op) {
-		case GET_REQUIRED_HW_INTERFACES:
-			flag = (CARD32*)ptr;
-			(*flag) = 0;
-			return TRUE;
-		default:
-			return FALSE;
-	}
+    xorgHWFlags *flag;
+    
+    switch (op) {
+	case GET_REQUIRED_HW_INTERFACES:
+	    flag = (CARD32*)ptr;
+	    (*flag) = 0;
+	    return TRUE;
+	default:
+	    return FALSE;
+    }
 }
 
 /*** Module and driver setup */
@@ -747,6 +879,7 @@ OMAPFBSetup(pointer module, pointer opts, int *errmaj, int *errmin)
 	if (!setupDone) {
 		setupDone = TRUE;
 		xf86AddDriver(&OMAPFB, module, HaveDriverFuncs);
+		LoaderRefSymLists(fbSymbols, NULL);
 		return (pointer)1;
 	} else {
 		if (errmaj) *errmaj = LDR_ONCEONLY;
